import pandas as pd
import numpy as np
from sklearn.metrics import mean_squared_error as mse
import time
import sys
import time
import pickle
import tensorflow as tf
from sklearn.model_selection import train_test_split


def simple_cov(_x, _y):
    return np.mean((_x-np.mean(_x))*(_y-np.mean(_y)), axis = 1)


def simple_covs(a, b):
    ac = a - a.mean(axis=-1, keepdims=True)
    bc = (b - b.mean(axis=-1, keepdims=True)) / b.shape[-1]
    return np.dot(ac, bc.T).T


def greedy_batch_acquizition(values, sample_size):
    return np.argsort(values)[::-1][:sample_size]


def update_learning_sets(X_train,
                         y_train,
                         X_pool,
                         y_pool,
                         ue_values,
                         acquizition = greedy_batch_acquizition,
                         sample_size = 10):
    inds = greedy_batch_acquizition(ue_values, sample_size)
    X_train = np.concatenate([X_train, X_pool[inds]])
    y_train = np.concatenate([y_train, y_pool[inds]])
    X_pool = np.delete(X_pool, inds, axis=0)
    y_pool = np.delete(y_pool, inds, axis=0)
    
    return X_train, y_train, X_pool, y_pool


def iterate_minibatches(inputs, targets, batchsize, shuffle=True):
    """
    Produces batches
    """
    assert len(inputs) == len(targets)
    if shuffle:
        indices = np.arange(len(inputs))
        np.random.shuffle(indices)
    if batchsize > len(inputs):
        yield inputs[indices], targets[indices]
    else:
        for start_idx in range(0, len(inputs) - batchsize + 1, batchsize):
            if shuffle:
                excerpt = indices[start_idx:start_idx + batchsize]
            else:
                excerpt = slice(start_idx, start_idx + batchsize)
            yield inputs[excerpt], targets[excerpt]
        if len(inputs) % batchsize > 0:
            if shuffle:
                excerpt = indices[start_idx+batchsize:len(inputs)]
            else:
                excerpt = slice(start_idx+batchsize, len(inputs))
            yield inputs[excerpt], targets[excerpt]

            
class NN:
    def __init__(
        self, ndim=1, layers=None, initializer=None, activation=None,
        random_state=42, dropout_layers=None
    ):
        if layers is None:
            layers = [32, 16]
        if initializer is None:
            initializer = tf.contrib.layers.xavier_initializer()
        if activation is None:
            activation = tf.nn.leaky_relu

        # for arg, value in locals().items():  # oh my god, what have i done
        #     if arg == 'self':
        #         continue
        #     self.__dict__[arg] = value

        if dropout_layers is None:
            dropout_layers = [True] * (len(layers)-1)

        self.input_data = tf.placeholder(tf.float32, [None, ndim])
        self.answer = tf.placeholder(tf.float32, [None, 1])
        self.keep_probability_ = tf.placeholder(tf.float32, name='keep_probability')
        self.keep_probability_inner_ = tf.placeholder(tf.float32, name='keep_probability')
        self.l2_reg_ = tf.placeholder(tf.float32, name='l2reg')
        
        W1 = tf.get_variable(name='W0', shape=(ndim, layers[0]), initializer=initializer)
        b1 = tf.get_variable(name='b0', shape=(layers[0],), initializer=initializer)

        h1 = activation(tf.matmul(self.input_data, W1) + b1)
        
        h_drop1 = tf.nn.dropout(h1, self.keep_probability_, noise_shape=[1, layers[0]])
        
        self.Ws = [W1]
        self.bs = [b1]
        self.hs = [h_drop1]
        
        for cnt_layer in range(1, len(layers)):
            self.Ws.append(
                tf.get_variable(
                    name = f'W{cnt_layer}',
                    shape = (layers[cnt_layer - 1], layers[cnt_layer]),
                    initializer = initializer))
            self.bs.append(
                tf.get_variable(
                    name = f'b{cnt_layer}',
                    shape = (layers[cnt_layer],),
                    initializer = initializer))
            h = activation(tf.matmul(self.hs[-1], self.Ws[-1]) + self.bs[-1])
            h = tf.nn.dropout(h,
                    self.keep_probability_ if dropout_layers[cnt_layer-1]
                                           else self.keep_probability_inner_,
                    noise_shape = [1, layers[cnt_layer]])
            self.hs.append(h)
                
        self.Ws.append(
            tf.get_variable(
                    name=f'W{len(layers)}',
                    shape=(layers[-1], 1),
                    initializer=initializer))
        self.bs.append(
            tf.get_variable(
                    name=f'b{len(layers)}',
                    shape=(1,),
                    initializer=initializer))
        self.output = activation(tf.matmul(self.hs[-1], self.Ws[-1]) + self.bs[-1])
        
        self.l2_regularizer = sum(tf.nn.l2_loss(Wxxx) for Wxxx in self.Ws) 
        self.mse = tf.losses.mean_squared_error(predictions = self.output, 
                                           labels = self.answer)
        self.loss = self.mse + self.l2_reg_*self.l2_regularizer
        self.train_step = tf.train.AdamOptimizer(learning_rate=1e-4).\
                                minimize(self.loss)
        
        # misc part for nngp eval
        self.KK_ = tf.placeholder(tf.float32)
        self.Q_ = tf.placeholder(tf.float32,
                                 [None, None])
        self.K_train_cov_inv_ = tf.placeholder(tf.float32, [None, None])
        self.gpue_tf = \
            tf.linalg.tensor_diag_part(self.KK_ - \
                                       tf.matmul(
                                           tf.matmul(
                                               tf.transpose(self.Q_), 
                                               self.K_train_cov_inv_), 
                                               self.Q_))
        
    def train(
        self, session, X_train, y_train, X_test, y_test, X_val, y_val,
        epochs=10000, early_stopping=True, validation_window=100,
        patience=3, keep_prob=1., l2_reg=0, verbose=True, batch_size=500,
    ):
        
        previous_error = 1e+10
        current_patience = patience
        
        for epoch_num in range(1, 1 + epochs):
            for X_batch, y_batch in iterate_minibatches(X_train, y_train, batch_size):
                session.run(self.train_step,
                            feed_dict={self.input_data: X_batch, 
                                       self.answer: y_batch, 
                                       self.keep_probability_inner_: keep_prob,
                                       self.keep_probability_: keep_prob, 
                                       self.l2_reg_: l2_reg})

            if (early_stopping) and epoch_num % validation_window == 0:
                rmse_train = np.sqrt(
                    session.run(self.mse, 
                             feed_dict={self.input_data: X_train, 
                                        self.answer: y_train, 
                                        self.keep_probability_inner_: 1.,
                                        self.keep_probability_: 1.}))
                rmse_test = np.sqrt(
                    session.run(self.mse, 
                             feed_dict={self.input_data: X_test, 
                                        self.answer: y_test, 
                                        self.keep_probability_inner_: 1.,
                                        self.keep_probability_: 1.}))       
                rmse_val = np.sqrt(
                    session.run(self.mse, 
                             feed_dict={self.input_data: X_val, 
                                        self.answer: y_val, 
                                        self.keep_probability_inner_: 1.,
                                        self.keep_probability_: 1.}))

                if rmse_val > previous_error:
                    current_patience -= 1
                else:
                    previous_error = rmse_val
                    current_patience = patience
                if verbose:
                    print(f'[{epoch_num}]'+\
                          f' RMSE train:{rmse_train:.3f}'+\
                          f' test:{rmse_test:.3f}'+\
                          f' val:{rmse_val:.3f}'+\
                          f' patience:{current_patience}')
                if current_patience <= 0:
                    if verbose:
                        print(f'No patience left at epoch {epoch_num}.'+\
                        ' Early stopping.')
                    break
        return epoch_num, rmse_train, rmse_test, rmse_val

    def predict(self, session, data, probability=1., probabitily_inner=1.):
        return session.run(self.output,
                           feed_dict={self.input_data: data,
                                      self.keep_probability_inner_: probabitily_inner,
                                      self.keep_probability_: probability})

    def get_mcd_UE(self, session, X_pool, nn_runs=25, probability=.5, use_inner=False):
        # nn inference
        mcd_realizations = np.zeros((X_pool.shape[0], nn_runs))
        for nn_run in range(nn_runs):
            mcd_realizations[:, nn_run] = \
                np.ravel(self.predict(session,
                                      data = X_pool,
                                      probability = probability,
                                      probabitily_inner = probability if use_inner
                                                          else 1.))
        return np.ravel(np.std(mcd_realizations, axis = 1))
    
    def get_nngp_UE(
        self, session, X_train, y_train, X_pool, y_pool, random_subsampling = -1,
        nn_runs=25, diag_eps=1e-6, inference_batch_size=1000, probability=.5,
        use_inner=False
    ):
        # data preparation
        train_len = len(X_train)
        if random_subsampling > 0:
            train_len = min(random_subsampling, train_len)
            random_train_inds = \
                np.random.permutation(range(train_len))[:random_subsampling]
            train_pool_samples = \
                np.concatenate([X_train[random_train_inds],
                                X_pool])
        else:
            train_pool_samples = \
                np.concatenate([X_train,
                                X_pool])
        # nn inference
        mcd_realizations = np.zeros((train_pool_samples.shape[0], nn_runs))
        for nn_run in range(nn_runs):
            mcd_realizations[:, nn_run] = \
                np.ravel(self.predict(session,
                                      data=train_pool_samples,
                                      probability=probability,
                                      probabitily_inner=probability if use_inner
                                                          else 1.))
        cov_matrix_train = np.cov(mcd_realizations[:train_len, :], 
                                  ddof = 0)
        cov_matrix_inv = np.linalg.inv(cov_matrix_train\
                                       + np.eye(train_len)*diag_eps)
        gp_ue = np.zeros((len(X_pool), ))
        
        cnt = 0 # for the case of inference_batch_size > len(X_pool)
        for cnt in range(len(X_pool) // inference_batch_size):
            left_ind = train_len + cnt*inference_batch_size
            right_ind = train_len + cnt*inference_batch_size+inference_batch_size
            pool_samples = mcd_realizations[left_ind:right_ind,:]
            Qs = simple_covs(mcd_realizations[:train_len,:], pool_samples).T
            KKs = np.var(pool_samples, axis=1)
            ws = session.run(self.gpue_tf, {
                             self.Q_: Qs,
                             self.K_train_cov_inv_: cov_matrix_inv,
                             self.KK_: KKs})
            gp_ue_currents = [0 if w < 0 else np.sqrt(w) for w in np.ravel(ws)]
            gp_ue[(left_ind - train_len):(right_ind - train_len)] = gp_ue_currents
        right_ind = train_len + cnt*inference_batch_size+inference_batch_size
        pool_samples = mcd_realizations[right_ind:,:]
        Qs = simple_covs(mcd_realizations[:train_len,:], pool_samples).T
        KKs = np.var(pool_samples, axis = 1)
        ws = session.run(self.gpue_tf, {
                             self.Q_: Qs,
                             self.K_train_cov_inv_: cov_matrix_inv,
                             self.KK_: KKs})
        gp_ue_currents = [0 if w < 0 else np.sqrt(w) for w in np.ravel(ws)]
        gp_ue[right_ind:] = gp_ue_currents
        return np.ravel(gp_ue)
        



import torch
from torch.nn.modules.loss import _Loss


class NLLRegLoss(_Loss):
    def __init__(self, size_average=None, reduce=None, reduction='mean', eps=1e-6):
        super(NLLRegLoss, self).__init__(size_average, reduce, reduction)
        self.eps = eps
    
    def forward(self, input, target):
        mu, sigma = input[:, 0], input[:, 1]
        target = target.reshape_as(mu)
        output = 0.5 * torch.log(sigma + self.eps) + 0.5 * (target - mu)**2 / sigma
        return getattr(output, self.reduction)()


from .mlp import MLP
from collections import OrderedDict
from itertools import count
import torch


class MLPEnsemble:
    def __init__(self, layers, n_models, reduction='mean', **kwargs):
        self.n_models = n_models
        self.models = [MLP(layers, **kwargs) for i in range(n_models)]
        self.reduction = reduction

    def fit(self, train_set, val_set, verbose=True, **kwargs):
        for i, model in enumerate(self.models): 
            if verbose:
                self._print_fit_status(i+1, self.n_models)
            model.fit(train_set, val_set, verbose=verbose, **kwargs)
    
    def state_dict(self):
        state_dict = OrderedDict({'{} model'.format(n): m.state_dict() 
                                  for n, m in zip(count(), self.models)})
        return state_dict
    
    def load_state_dict(self, state_dict):
        for n, m in enumerate(self.models):
            m.load_state_dict(state_dict['{} model'.format(n)])
    
    def train(self):
        [m.train() for m in self.models]
        
    def eval(self):
        [m.eval() for m in self.models]
    
    def __call__(self, x, reduction='default', **kwargs):
        if 'dropout_mask' in kwargs and isinstance(kwargs['dropout_mask'], list):
            masks = kwargs.pop('dropout_mask')
            res = torch.stack([m(x, dropout_mask = dpm, **kwargs) for m, dpm in zip(self.models, masks)])
        else:
            res = torch.stack([m(x, **kwargs) for m in self.models])

        if reduction == 'default':
            reduction = self.reduction

        if reduction is None:
            res = res
        elif reduction == 'mean':
            res = res.mean(dim=0)
        elif reduction == 'nll':
            means = res[:, :, 0]
            sigmas = res[:, :, 1]
            res = torch.stack([means.mean(dim=0), sigmas.mean(dim=0) +
                               (means**2).mean(dim=0) - means.mean(dim=0)**2], dim=1)
        return res
    
    def _print_fit_status(self, n_model, n_models):
        print('Fit [{}/{}] model:'.format(n_model, n_models))



import os
from pathlib import Path

import tarfile

from experiments.experiment_setup import ROOT_DIR
from .saver import DataSaver
from .downloader import download


URL = 'http://yaroslavvb.com/upload/notMNIST/notMNIST_large.tar.gz'


class NotMNISTData:
    def __init__(self, use_cache=False, val_split=0.2):
        raise NotImplementedError

        self.use_cache = use_cache
        cache_dir = Path(ROOT_DIR)/'dataloader'/'data'/'not_mnist'
        self.saver = DataSaver(cache_dir)
        self.val_split = val_split
        self._build_dataset(cache_dir)

    def dataset(self, label):
        if self.use_cache:
            return self.saver.load(label)

        data = self.data[label]
        x, y = data[:, 1:], data[:, :1]
        self.saver.save(x, y, label)
        return x, y

    def _build_dataset(self, cache_dir):
        data_path = download(cache_dir, 'notMNIST_small.tar.gz', URL)
        import ipdb; ipdb.set_trace()
        with tarfile.open(data_path, "r:gz") as tar_ref:
            tar_ref.extractall(cache_dir)

        base_dir = cache_dir / 'notMNIST_small'

        class_dirs = sorted(os.listdir(base_dir))
        # with zipfile.ZipFile(data_path, 'r') as zip_ref:
        #     zip_ref.extractall(cache_dir)

        # train, val = train_test_split(table, test_size=self.val_split, shuffle=True)
        # self.data = {
        #     'train': train,
        #     'val': val,
        #     'all': np.concatenate((train, val))
        # }


if __name__ == '__main__':
    dataset = NotMNISTData()
    x_train, y_train = dataset.dataset('train')
    x_val, y_val = dataset.dataset('val')
    print(x_train.shape, y_train.shape, y_val.shape)
    print(x_train[:5])
    print(y_train[:5])


import torch
from torch import nn
import torch.nn.functional as F

from torch.hub import load_state_dict_from_url
from torchvision.models.resnet import ResNet, BasicBlock, conv3x3, conv1x1, Bottleneck
# from torchvision.models.resnet import resnet18


model_urls = {
    'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
}


class ResNetMasked(ResNet):
    def forward(self, x, dropout_rate=0.5, dropout_mask=None):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        if dropout_mask is not None:
            x = x * dropout_mask(x, dropout_rate, 0)
        x = self.fc(x)

        return x


class ResNetLinear(ResNet):
    def forward(self, x, dropout_rate=0.5, dropout_mask=None):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = F.celu(self.fc(x))
        if dropout_mask is not None:
            x = x * dropout_mask(x, dropout_rate, 0)
        else:
            x = F.dropout(x, self.dropout_rate)
        x = F.celu(self.fc2(x))
        x = F.dropout(x, self.dropout_rate)
        x = self.fc3(x)

        return x


def resnet_linear(pretrained=False, dropout_rate=0.5, freeze=False):
    base = resnet18(pretrained=pretrained, mode='linear')
    base.dropout_rate = dropout_rate
    base.fc = nn.Linear(512, 512)
    base.fc2 = nn.Linear(512, 256)
    base.fc3 = nn.Linear(256, 10)

    if freeze:
        for param in list(base.parameters())[:-8]:
            param.requires_grad = False

    return base


def resnet_masked(pretrained=False):
    base = resnet18(pretrained=pretrained)
    base.fc = nn.Linear(512, 10)

    return base


def _resnet(arch, block, layers, pretrained, progress, mode='masked', **kwargs):
    if mode == 'masked':
        model = ResNetMasked(block, layers, **kwargs)
    else:
        model = ResNetLinear(block, layers, **kwargs)
    if pretrained:
        state_dict = load_state_dict_from_url(model_urls[arch],
                                              progress=progress)
        model.load_state_dict(state_dict)
    return model


def resnet18(pretrained=False, progress=True, **kwargs):
    r"""ResNet-18 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet('resnet18', BasicBlock, [2, 2, 2, 2], pretrained, progress,
                   **kwargs)




import os
import os.path as path
import zipfile
import pandas as pd
import wget


class BlogFeedbackData:
    def __init__(
            self, cache_dir='dataloader/data/blog_feedback',
            url='https://archive.ics.uci.edu/ml/machine-learning-databases/00304/BlogFeedback.zip'):
        self.cache_dir = cache_dir
        self.url = url
        self.filename = 'bf.zip'

    def dataset(self, label='train'):
        try:
            return self._parse(label)
        except FileNotFoundError:
            self._load()
            self._parse(label)

    def _parse(self, label):
        if label == 'train':
            train_file = 'blogData_train.csv'
            df = pd.read_csv(os.path.join(self.cache_dir, train_file), header=None)
        elif label == 'test':
            files = os.listdir(self.cache_dir)
            test_files = [file for file in files if file.startswith('blogData_test')]
            test_files = [os.path.join(self.cache_dir, file) for file in test_files]
            df = pd.concat((pd.read_csv(f, header=None) for f in test_files))
        else:
            raise RuntimeError("Wrong label")

        return df.loc[:, :279].values, df.loc[:, 280].values

    def _load(self):
        if not path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)
        zip_path = path.join(self.cache_dir, self.filename)
        if not path.exists(zip_path):
            wget.download(self.url, zip_path)
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            zip_ref.extractall(self.cache_dir)


if __name__ == '__main__':
    print(BlogFeedbackData().dataset())



from .sample_selector import EagerSampleSelector
from .oracle import IdentityOracle


class ALTrainer:
    """
    Active Learning trainer

    trains on train data
    on each iteration extends training sets from sampling the pool
    """
    def __init__(
            self, model, estimator, oracle=None, sampler=None, y_pool=None, iterations=10,
            update_size=100, verbose=True, patience=10, val_on_pool=False):
        self.model = model
        self.estimator = estimator
        self.sampler = sampler
        self.oracle = oracle
        self.iterations = iterations
        self.update_size = update_size
        self.verbose = verbose
        self.patience = patience
        self.val_on_pool = val_on_pool

        if sampler is None:
            sampler = EagerSampleSelector()
        self.sampler = sampler

        if oracle is None:
            if y_pool is not None:
                oracle = IdentityOracle(y_pool)
            else:
                raise ValueError("Either oracle or y_pool should be initialized")
        self.oracle = oracle

    def train(self, x_train, y_train, x_val, y_val, x_pool):
        errors = []

        for al_iteration in range(self.iterations):
            print("Iteration", al_iteration+1)
            print(x_pool.shape)
            # retrain net
            self.model.fit((x_train, y_train), (x_val, y_val), verbose=self.verbose, patience=self.patience)
            if self.val_on_pool:
                error = self.model.evaluate((x_pool, self.oracle.y_set))
            else:
                error = self.model.evaluate((x_val, y_val))

            print('Validation error after training: %.3f' % error)
            errors.append(error)

            # update pool
            if hasattr(self.estimator, 'reset'):
                self.estimator.reset()
            uncertainties = self.estimator.estimate(x_pool, x_train)
            x_train, y_train, x_pool = self.sampler.update_sets(
                x_train, y_train, x_pool, uncertainties, self.update_size, self.oracle
            )

            if self.verbose:
                print('Uncertainties', uncertainties[:20])
                print('Top uncertainties', uncertainties[uncertainties.argsort()[-10:][::-1]])

        return errors




import torch
import numpy as np


class NNGPRegression:
    def __init__(self, net, nn_runs=25, diag_eps=1e-6, dropout_rate=0.5):
        self.net = net
        self.nn_runs = nn_runs
        self.diag_eps = diag_eps
        self.dropout_rate = dropout_rate

    def estimate(self, x_pool, x_train):
        train_pool_samples = np.concatenate([x_train, x_pool])
        train_len = len(x_train)

        mcd_predictions = self._mcd_predict(train_pool_samples)

        # covariance matrix with regularization
        cov_matrix_train = np.cov(mcd_predictions[:train_len, :], ddof=0)
        cov_matrix_inv = np.linalg.inv(cov_matrix_train + np.eye(train_len)*self.diag_eps)

        pool_samples = mcd_predictions[train_len:]
        Qs = self.simple_covs(mcd_predictions[:train_len, :], pool_samples).T
        Qt_K_Q = np.matmul(np.matmul(Qs.T, cov_matrix_inv), Qs)
        KKs = np.var(pool_samples, axis=1)

        ws = KKs - np.diag(Qt_K_Q)

        gp_ue = [0 if w < 0 else np.sqrt(w) for w in np.ravel(ws)]
        return np.ravel(gp_ue)

    def _mcd_predict(self, train_pool_samples):
        mcd_predictions = np.zeros((train_pool_samples.shape[0], self.nn_runs))
        with torch.no_grad():
            for nn_run in range(self.nn_runs):
                prediction = self.net(train_pool_samples, dropout_rate=self.dropout_rate).to('cpu')
                mcd_predictions[:, nn_run] = np.ravel(prediction)
        return mcd_predictions

    @staticmethod
    def simple_covs(a, b):
        ac = a - a.mean(axis=-1, keepdims=True)
        bc = (b - b.mean(axis=-1, keepdims=True)) / b.shape[-1]
        return np.dot(ac, bc.T).T




import numpy as np
import torch  
    
class EnsembleMCDUE:
    """
    Estimate uncertainty for samples with Ensemble and MCDUE approach
    Propose that ensemble contains nets with one output
    """ 
    def __init__(self, net, nn_runs=5, dropout_rate=.5):
        self.net = net
        self.nn_runs = nn_runs
        self.dropout_rate = dropout_rate
        self.n_models = self.net.n_models
    
    def estimate(self, X_pool, **kwargs):
        mcd_realizations = []

        with torch.no_grad():
            for nn_run in range(self.nn_runs):
                prediction = self.net(X_pool, reduction=None, dropout_rate=self.dropout_rate, **kwargs)
                prediction = prediction.to('cpu')
                mcd_realizations.append(prediction)
        
        mcd_realizations = torch.cat(mcd_realizations, dim=0)
        return np.ravel(mcd_realizations.std(dim=0, unbiased=False))
    
    
class EnsembleMCDUEMasked:
    """
    Estimate uncertainty for samples with Ensemble and MCDUE approach
    """
    def __init__(self, net, nn_runs=25, dropout_rate=.5, dropout_mask=None):
        self.net = net
        self.nn_runs = nn_runs
        self.dropout_rate = dropout_rate
        self.dropout_mask = dropout_mask

    def estimate(self, X_pool, *args):
        mcd_realizations = np.zeros((X_pool.shape[0], self.nn_runs))
        mcd_realizations = []
        
        with torch.no_grad():
            # Some mask needs first run without dropout, i.e. decorrelation mask
            if hasattr(self.dropout_mask, 'dry_run') and self.dropout_mask.dry_run:
                self.net(X_pool, dropout_rate=self.dropout_rate, dropout_mask=self.dropout_mask)

            # Get mcdue estimation
            for nn_run in range(self.nn_runs):
                prediction = self.net(X_pool, reduction=None,
                                      dropout_rate=self.dropout_rate,
                                      dropout_mask=self.dropout_mask)
                prediction = prediction.to('cpu')
                mcd_realizations.append(prediction)
                
        mcd_realizations = torch.cat(mcd_realizations, dim=0)
        return np.ravel(mcd_realizations.std(dim=0, unbiased=False))
    
    
class EnsembleUE(EnsembleMCDUE):
    """
    Estimate uncertainty for samples with Ensemble approach.
    Propose that ensemble contains nets with one output
    """
    def __init__(self, net):
        super(EnsembleUE, self).__init__(net, nn_runs=1, dropout_rate=0)
        
             
class EnsembleNLLUE():
    """
    Estimate uncertainty for samples with Ensemble approach.
    Ensemble must contains nets with two outputs: mean and sigma_squared
    """
    def __init__(self, net):
        self.net = net
        
    def estimate(self, X_pool, **kwargs):
        with torch.no_grad():
            res = self.net(X_pool, reduction='nll')
            sigma = res[:, 1].to('cpu')
            sigma = np.sqrt(sigma)
            
        return np.ravel(sigma)


from os import path

import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
import zipfile

from .config import DATA_DIR
from .saver import DataSaver
from .downloader import download


URL = 'http://archive.ics.uci.edu/ml/machine-learning-databases/00316/UCI%20CBM%20Dataset.zip'


class NavalPropulsionData:
    """Load/provides naval propulsion dataset"""
    def __init__(self, use_cache=False, val_split=0.2):
        self.use_cache = use_cache
        cache_dir = path.join(DATA_DIR, 'dataloader/data/naval_propulsion')
        self.saver = DataSaver(cache_dir)
        self.val_split = val_split
        self._build_dataset(cache_dir)

    def dataset(self, label):
        if self.use_cache:
            return self.saver.load(label)

        data = self.data[label]
        x, y = data[:, :-2], data[:, -1:]
        self.saver.save(x, y, label)
        return x, y

    def _build_dataset(self, cache_dir):
        data_path = download(cache_dir, 'naval_propulsion.zip', URL)
        with zipfile.ZipFile(data_path, 'r') as zip_ref:
            zip_ref.extractall(cache_dir)
        file_path = path.join(cache_dir, 'UCI CBM Dataset', 'data.txt')
        self.df = pd.read_csv(file_path, delim_whitespace=True, header=None)
        table = self.df.to_numpy()
        if self.val_split != 0:
            train, val = train_test_split(table, test_size=self.val_split, shuffle=True)
        else:
            train, val = table, []
        self.data = {
            'train': train,
            'val': val,
            'all': np.concatenate((train, val))
        }


if __name__ == '__main__':
    dataset = NavalPropulsionData()
    x_train, y_train = dataset.dataset('train')
    x_val, y_val = dataset.dataset('val')
    print(x_train.shape, y_val.shape)




import os
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


PROJECT_ROOT = os.path.abspath(os.path.dirname(__file__))


def load_requirements(filename):
    with open(os.path.join(PROJECT_ROOT, filename), "r") as f:
        return f.read().splitlines()


setuptools.setup(
    name="alpaca-ml",
    version="0.0.5",
    author="Maxim Panov and Evgenii Tsymbalov and Kirill Fedyanin",
    author_email="k.fedyanin@skoltech.ru",
    description="Active learning utilities for machine learning applications",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/stat-ml/alpaca",
    packages=setuptools.find_packages(),
    license="Apache License 2.0",
    install_requires=load_requirements("requirements.txt"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
)


import numpy as np
from math import log2
from scipy.stats import percentileofscore


def get_uq_metrics(estimations, errors, acc_percentile=0.1, **kwargs):
    acc = uq_accuracy(estimations, errors, acc_percentile)
    ndcg = uq_ndcg(errors, estimations, **kwargs)
    ll = uq_ll(errors, estimations)
    return acc, ndcg, ll


def uq_accuracy(uq, errors, percentile=0.1):
    """Shows intersection of worst by error/uq in percentile"""
    k = int(len(uq)*percentile)
    worst_uq = np.argsort(np.ravel(uq))[-k:]
    worst_error = np.argsort(np.ravel(errors))[-k:]
    return len(set(worst_uq).intersection(set(worst_error)))/k


def dcg(relevances, scores, k):
    """
    Discounting cumulative gain, metric of ranking quality
    For UQ - relevance is ~ error, scores is uq
    """
    relevances = np.ravel(relevances)
    scores = np.ravel(scores)

    ranking = np.argsort(scores)[::-1]
    metric = 0
    for rank, score_id in enumerate(ranking[:k]):
        metric += relevances[score_id] / log2(rank+2)

    return metric


def ndcg(relevances, scores):
    """
    Normalized DCG. We norm fact DCG on ideal DCG score
    expect relevances, scores to be numpy ndarrays
    """
    k = sum(relevances != 0)
    return dcg(relevances, scores, k) / dcg(relevances, relevances, k)


def uq_ndcg(errors, uq, bins=None):
    """
    In UQ we care most of top erros,
    so we restructure errors to give top errors bigger relevance
    """
    if bins is None:
        bins = [80, 95, 99]

    sorted_errors = sorted(errors)
    errors_percentiles = [percentileofscore(sorted_errors, error) for error in errors]
    errors_digitized = np.digitize(errors_percentiles, bins)

    return ndcg(errors_digitized, uq)


def uq_ll(errors, uq):
    errors = np.ravel(errors)
    uq_squared = np.square(np.ravel(uq)) + 1e-10
    return -np.mean(np.log(uq_squared)/2 + np.square(errors)/2/uq_squared)



import torch
import torch.nn as nn
from torch.nn import functional as F
from alpaca.dataloader import loader


class AutoEncoder(nn.Module):
    def __init__(self, input_size, hidden_size, embedding_size, lr=1e-3, dropout_rate=0.3):
        super(AutoEncoder, self).__init__()

        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.input_size = input_size

        self.fc1 = nn.Linear(input_size, hidden_size)
        self.fc2 = nn.Linear(hidden_size, embedding_size)
        self.fc3 = nn.Linear(embedding_size, hidden_size)
        self.fc4 = nn.Linear(hidden_size, input_size)

        self.to(self.device)
        self.optimizer = torch.optim.Adam(self.parameters(), lr=lr)
        self.double()

    def encode(self, x):
        h1 = F.leaky_relu(self.fc1(x))
        return self.fc2(h1)

    def decode(self, z):
        h3 = F.relu(self.fc3(z))
        return self.fc4(h3)
        # return 3*torch.tanh(self.fc4(h3))

    def predict(self, x):
        self.eval()
        x = torch.DoubleTensor(x).to('cuda')
        return self(x).cpu().detach().numpy()

    def forward(self, x):
        encoded = self.encode(x.view(-1, self.input_size))
        decoded = self.decode(encoded)
        return decoded

    def fit(self, x_train, batch_size=128):
        self.train()
        train_loss = 0
        train_loader = loader(x_train, x_train, batch_size=batch_size)
        for batch_idx, (data, _) in enumerate(train_loader):
            data = data.to(self.device)
            self.optimizer.zero_grad()
            recon_batch = self(data)
            loss = F.mse_loss(recon_batch, data)
            loss.backward()
            train_loss += loss.item()
            self.optimizer.step()
        return train_loss/len(x_train)

    def evaluate(self, x_val):
        self.eval()
        test_loss = 0
        test_loader = loader(x_val, x_val)
        with torch.no_grad():
            for i, (data, _) in enumerate(test_loader):
                data = data.to(self.device)
                recon_batch = self(data)
                test_loss += F.mse_loss(recon_batch, data).item()

        test_loss /= len(test_loader.dataset)
        return test_loss


class VAE(nn.Module):
    def __init__(self, input_size, hidden_size, embedding_size):
        super(VAE, self).__init__()

        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.input_size = input_size

        self.fc1 = nn.Linear(input_size, hidden_size)
        self.fc21 = nn.Linear(hidden_size, embedding_size)
        self.fc22 = nn.Linear(hidden_size, embedding_size)
        self.fc3 = nn.Linear(embedding_size, hidden_size)
        self.fc4 = nn.Linear(hidden_size, input_size)

        self.to(self.device)
        self.optimizer = torch.optim.Adam(self.parameters(), lr=1e-3)
        self.double()

    def encode(self, x):
        h1 = F.leaky_relu(self.fc1(x))
        return self.fc21(h1), self.fc22(h1)

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5*logvar)
        eps = torch.randn_like(std)
        return mu + eps*std

    def decode(self, z):
        h3 = F.relu(self.fc3(z))
        return torch.tanh(self.fc4(h3))

    def forward(self, x):
        mu, logvar = self.encode(x.view(-1, self.input_size))
        z = self.reparameterize(mu, logvar)
        return self.decode(z), mu, logvar

    # Reconstruction + KL divergence losses summed over all elements and batch
    def _loss_function(self, recon_x, x, mu, logvar):
        BCE = F.binary_cross_entropy(recon_x, x.view(-1, self.input_size), reduction='sum')
        # see Appendix B from VAE paper:
        # Kingma and Welling. Auto-Encoding Variational Bayes. ICLR, 2014
        # https://arxiv.org/abs/1312.6114
        # 0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)
        KLD = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())

        return BCE + KLD

    def fit(self, x_train, batch_size=128):
        self.train()
        train_loss = 0
        train_loader = loader(x_train, x_train, batch_size=batch_size)
        for batch_idx, (data, _) in enumerate(train_loader):
            data = data.to(self.device)
            self.optimizer.zero_grad()
            recon_batch, mu, logvar = self(data)
            loss = self._loss_function(recon_batch, data, mu, logvar)
            loss.backward()
            train_loss += loss.item()
            self.optimizer.step()

    def evaluate(self, x_val):
        self.eval()
        test_loss = 0
        test_loader = loader(x_val, x_val)
        with torch.no_grad():
            for i, (data, _) in enumerate(test_loader):
                data = data.to(self.device)
                recon_batch, mu, logvar = self(data)
                test_loss += self._loss_function(recon_batch, data, mu, logvar).item()

        test_loss /= len(test_loader.dataset)
        return test_loss

    def predict(self, x_batch):
        with torch.no_grad():
            result = self(torch.from_numpy(x_batch).to(self.device))
        return result[0].tolist()




import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np


class Dense(nn.Module):
    def __init__(self, layer_sizes, postprocessing=lambda x: x, activation=None, dropout_rate=0):
        super().__init__()

        if activation is None:
            self.activation = F.celu
        else:
            self.activation = activation

        self.layer_sizes = layer_sizes
        self.fcs = []
        for i, layer in enumerate(layer_sizes[:-1]):
            fc = nn.Linear(layer, layer_sizes[i + 1])
            setattr(self, 'fc' + str(i), fc)  # to register params
            self.fcs.append(fc)
        self.postprocessing = postprocessing

        self.dropout = nn.Dropout(dropout_rate)
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.to(device)

    def forward(self, x, dropout_rate=0.5, dropout_mask=None):
        out = self.activation(self.fcs[0](x))

        for layer_num, fc in enumerate(self.fcs[1:-1]):
            out = self.activation(fc(out))
            # out = self.dropout(out)

        if dropout_mask is not None:
            out = out * dropout_mask(out, dropout_rate, layer_num=0)
        else:
            out = self.dropout(out)

        out = self.fcs[-1](out)
        return out
